// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
// File        : vmadump.c
// Description : Dump the shared vma of a process
// License     :
//
//  Copyright (C) 2020 Rachid Koucha <rachid dot koucha at gmail dot com>
//
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Evolutions  :
//
//     10-Sep-2020 R. Koucha      - Creation
//
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=


#include <assert.h>
#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#include <libgen.h>
#include <string.h>
#include <ctype.h>
#include <stddef.h>
#include <sys/sendfile.h>



//---------------------------------------------------------------------------
// Name  : dbg_level
// Usage : Debug level
//---------------------------------------------------------------------------
static int dbg_level;

#define DBG(level, fmt, ...) do {		              \
    if (level <= dbg_level) {                                 \
      fprintf(stderr,                                         \
              "DBG_%d (%s#%d): "fmt,                          \
	      level, __FUNCTION__, __LINE__, ## __VA_ARGS__); \
    }                                                         \
  } while(0)


//---------------------------------------------------------------------------
// Name  : ERR
// Usage : Error display
//---------------------------------------------------------------------------
#define ERR(fmt, ...) do {                            \
    fprintf(stderr,                                   \
            "ERROR@%s#%d: "fmt,                       \
             __FUNCTION__, __LINE__, ## __VA_ARGS__); \
                         } while(0)



//---------------------------------------------------------------------------
// Name  : vma_t
// Usage : VMA descriptor
//---------------------------------------------------------------------------
typedef struct link
{
  struct link *next; // Following record in the ordered list of VMA
  struct link *prev; // Previous record in the ordered list of VMA
} link_t;

typedef struct vma_desc
{
  char   *fname;       // File name
  int     fd;          // File descriptor
  char   *start_addr;  // VMA's starting address (used to sort the records)
  char   *end_addr;    // VMA's ending address
  size_t  len;         // VMA length
  char   *addr;        // Mapping address in current process

  link_t links;
} vma_t;

static link_t vma_list;

#define LINK2VMA(l) ((vma_t *)((char *)(l) - offsetof(vma_t, links)))



//---------------------------------------------------------------------------
// Name  : VMAD_META_...
// Usage : Indexes of the meta-characters (wildcards)
//---------------------------------------------------------------------------
#define VMAD_META_0_OR_MORE_CHARS          '*'
#define VMAD_META_1_OR_MORE_CHARS          '+'
#define VMAD_META_0_OR_1_CHAR              '?'
#define VMAD_META_ANY_SINGLE_CHAR          '.'
#define VMAD_META_0_OR_MORE_DIGITS         '$'
#define VMAD_META_1_OR_MORE_DIGITS         '!'
#define VMAD_META_0_OR_1_DIGIT             '%'
#define VMAD_META_ANY_SINGLE_DIGIT         ':'
#define VMAD_META_INHIBITION               '\\'


#define vmad_isdigit(c) (((c) >= '0') && ((c) <= '9'))

static char vmad_meta[128] = {

  [VMAD_META_0_OR_MORE_CHARS]  = 1,
  [VMAD_META_1_OR_MORE_CHARS]  = 1,
  [VMAD_META_0_OR_1_CHAR]      = 1,
  [VMAD_META_ANY_SINGLE_CHAR]  = 1,
  [VMAD_META_0_OR_MORE_DIGITS] = 1,
  [VMAD_META_1_OR_MORE_DIGITS] = 1,
  [VMAD_META_0_OR_1_DIGIT]     = 1,
  [VMAD_META_ANY_SINGLE_DIGIT] = 1,
  [VMAD_META_INHIBITION]       = 1

};

#define vmad_is_wildcard(p) ((((size_t)(p)) < sizeof(vmad_meta)) && \
                           vmad_meta[(size_t)(p)])



// ----------------------------------------------------------------------------
// Name   : vmad_match
// Usage  : Compare 2 strings taking in account the meta-characters:
//
//             . '*' 0 or more chars
//             . '+' 1 or more chars
//             . '?' 0 or 1 char
//             . '.' 1 char
//
//             . '$' 0 or more digits
//             . '!' 1 or more digits
//             . '%' 0 or 1 digit
//             . ':' 1 digit
//
//             . '\' inhibit following char
//
// return :  0,  if pattern matches 's'
//          -1,  if no match
// ----------------------------------------------------------------------------
static int vmad_match(
                    const char *pattern,
                    const char *s
                   )
{

  // Avoid useless recursive calls
  while (!vmad_is_wildcard(*pattern)) {

    // Pattern is not a wildcard
    // ==> Compare two non wildcard chars
    if (*pattern == *s) {

      // If end of both strings, matching is OK
      if (!(*pattern) && !(*s)) {

        return 0;

      }

      pattern ++;
      s ++;

    } else {

      // Matching failed
      return -1;

    } // End if comparison OK

  } // End while

  // Here pattern is a wildcard (so it is not NUL !)

  switch(*pattern) {

    case VMAD_META_0_OR_MORE_CHARS: {

      // Try 0 char
      if (0 == vmad_match(pattern + 1, s)) {

        return 0;

      } else {

        if (*s) {

          // Try more chars
	  return vmad_match(pattern, s + 1);

        } else {

          // End of 's' ==> 'pattern' is longer than 's' 
	  return -1;

        }

      }

    }
    break;
  
    case VMAD_META_1_OR_MORE_CHARS: {

      // Try 1 char
      if (*s) {

        if (0 == vmad_match(pattern + 1, s + 1)) {

          return 0;

        } else {

          // Try more chars
	  return vmad_match(pattern, s + 1);

        }

      } else {

        // End of 's' but 'pattern' is not empty
        return -1;

      }
    }
    break;

    case VMAD_META_0_OR_1_CHAR: {

      // Try 0 char
      if (0 == vmad_match(pattern + 1, s)) {

        return 0;

      } else {

        if (*s) {

          // Try 1 char
	  return vmad_match(pattern + 1, s + 1);

        } else {

          // End of 's' but 'pattern' is not empty
	  return -1;

        }

      }

    }
    break;

    case VMAD_META_ANY_SINGLE_CHAR: {

      if (*s) {

        return vmad_match(pattern + 1, s + 1);

      } else {

        return -1;

      }
    }
    break;

    case VMAD_META_0_OR_MORE_DIGITS: {

      // Try 0 digit
      if (0 == vmad_match(pattern + 1, s)) {

        return 0;

      } else {

        if (vmad_isdigit(*s)) {

          // Try more digits
	  return vmad_match(pattern, s + 1);

        } else {

          // End of 's': 'pattern' is longer than 's' or '*s' is not a digit 
	  return -1;

        }
      }
    }
    break;

    case VMAD_META_1_OR_MORE_DIGITS: {

      // Try 1 digit
      if (vmad_isdigit(*s)) {

        if (0 == vmad_match(pattern + 1, s + 1)) {

          return 0;

        } else {

          // Try more digits
	  return vmad_match(pattern, s + 1);

        }

      } else {

        // End of 's' but 'pattern' is not empty or '*s' is not a digit
        return -1;

      }
    }
    break;

    case VMAD_META_0_OR_1_DIGIT: {

      // Try 0 digit
      if (0 == vmad_match(pattern + 1, s)) {

        return 0;

      } else {

        if (vmad_isdigit(*s)) {

          // Try 1 digit
	  return vmad_match(pattern + 1, s + 1);

        } else {

          // End of 's' but 'pattern' is not empty or '*s' is not a digit
	  return -1;

        }
      }
    }
    break;

    case VMAD_META_ANY_SINGLE_DIGIT: {

      if (vmad_isdigit(*s)) {

        return vmad_match (pattern + 1, s + 1);

      } else {

        return -1;

      }
    }
    break;

    case VMAD_META_INHIBITION: {

      pattern++;

      if (*pattern == *s) {

        if (*pattern) {

          return vmad_match(pattern + 1, s + 1);

        } else {

          // Both strings are finished
          return 0;

        }

      }

      return -1;
    }
    break;

    default: {

      assert(0);

    }

  } // End switch

} // vmad_match


// ----------------------------------------------------------------------------
// Name   : vmad_map_pid
// Usage  : Map VMAs matching patterns in process#pid
// return :  0,  if OK
//          -1,  upon error
// ----------------------------------------------------------------------------
int vmad_map_pid(
               pid_t        pid,
               const char **pattern,
               int          nb_patterns,
               int         *nb_vma
              )
{
FILE   *f;
char    fname[256];
int     rc;
char    line[1024];
char   *p;
size_t  l;
char   *fields[6];
#define ADDR_FIELD   0
#define PERM_FIELD   1
#define OFFSET_FIELD 2
#define DEV_FIELD    3
#define INO_FIELD    4
#define FNAME_FIELD  5
int     i, j;
vma_t  *vma;
link_t *next;

  *nb_vma = 0;

  rc = snprintf(fname, sizeof(fname), "/proc/%d/maps", pid);
  if (rc < 0 || rc >= (int)sizeof(fname)) {

    if (rc >= 0) {
      errno = ENOSPC;
    }
    ERR("snprintf(): '%m' (%d)\n", errno);
    return -1;

  }

  f = fopen(fname, "r");
  if (!f) {

    ERR("fopen(%s): '%m' (%d)\n", fname, errno);
    return -1;

  }

  while (fgets(line, sizeof(line), f)) {

    l = strlen(line);

    // Remove the terminating newline if any
    if (line[l - 1] == '\n') {

      line[l - 1] = '\0';

    } else {

      // Every line in this file is terminated by a newline
      // So, here this means that the buffer is too small
      errno = ENOSPC;
      ERR("Internal buffer too small to read lines in '%s'!\n", fname);
      goto err;

    }

    // Get the fields
    p = line;
    for (i = 0; i < FNAME_FIELD; i ++) {

      // Field
      fields[i] = p;
      while (!isspace(*p)) {
        p ++;
      } // End while
      *p = '\0';
      p ++;
      while (isspace(*p)) {
        p ++;
      } // End while

    } // End for

    // Last field
    fields[i] = p;

    // Match the file name
    for (j = 0; j < nb_patterns; j ++) {

      if (0 == vmad_match(pattern[j], fields[FNAME_FIELD])) {

        DBG(4, "File name '%s' matched pattern '%s'\n",
            fields[FNAME_FIELD], pattern[j]);

        break;
      }

    } // End for

    // If the file name matched a pattern
    if (j < nb_patterns) {

      // Make sure that it is a shared segment
      p = fields[PERM_FIELD];
      if (p[3] != 's') {

        errno = EINVAL;
        ERR("VMA of file '%s' is not shared (permission = '%s')!\n",
            fields[FNAME_FIELD],
            fields[PERM_FIELD]);
        goto err;

      }

      // One more entry in the VMA table
      (*nb_vma) += 1;
      vma = (vma_t *)malloc(sizeof(vma_t));
      if (!vma) {

        ERR("realloc(): '%m' (%d)\n", errno);
        goto err;

      }

      // Starting/ending VMA addresses
      p = fields[ADDR_FIELD];
      while (*p != '-') {

        p ++;

      }
      *p = '\0';

      // Populate the VMA entry
      vma->fname = strdup(fields[FNAME_FIELD]);
      if (!(vma->fname)) {

        ERR("strdup(%s): '%m' (%d)\n", fields[FNAME_FIELD], errno);
        goto err;

      }
      vma->start_addr = (char *)strtoul(fields[ADDR_FIELD], NULL, 16);
      vma->end_addr = (char *)strtoul(p + 1, NULL, 16);
      vma->len = vma->end_addr - vma->start_addr;

      // Insert the record in the sorted list
      next = vma_list.next;
      while ((next != &vma_list) && (LINK2VMA(next)->start_addr < vma->start_addr)) {

        next = next->next;

      } // End while

      vma->links.next = next;
      vma->links.prev = next->prev;
      next->prev->next = &(vma->links);
      next->prev = &(vma->links);

      DBG(4, "New VMA: '%s', %p-%p (%zu bytes)\n",
          fields[FNAME_FIELD],
          vma->start_addr,
          vma->end_addr,
          vma->len
         );

    } // End if file name matched

  } // End while

  (void)fclose(f);

  return 0;

err:

  {
  int err_sav = errno;

    if (f) {
      (void)fclose(f);
    }

    errno = err_sav;

  }

  return -1;

} // vmad_map_pid



// ----------------------------------------------------------------------------
// Name   : vmad_long_ops
// Usage  : Option on the command line
// ----------------------------------------------------------------------------
static struct option vmad_long_opts[] =
{

  { "size",    required_argument, NULL, 's' },
  { "offset",  required_argument, NULL, 'o' },
  { "pattern", required_argument, NULL, 'p' },
  { "pid",     required_argument, NULL, 'P' },
  { "debug",   required_argument, NULL, 'd' },
  { "force",   no_argument,       NULL, 'F' },
  { "help",    no_argument,       NULL, 'h' },

  // Last entry
  { NULL,      0,                 NULL, 0   }

};


// ----------------------------------------------------------------------------
// Name   : vmad_usage
// Usage  : Program's help
// Return : None
// ----------------------------------------------------------------------------
static void vmad_usage(char *pname)
{

  fprintf(stderr,
          "Usage: %s [-size|--size=size] [-o|--offset=off] -p|--pattern=pattern\n"
          "               -P|--pid=pid [-d|--debug=level] [-F] [-h|--help]\n"
          "\n"
          "   -s|--size=size      : Number of bytes to dump\n"
          "                         (default all the mapping from offset)\n"
          "   -o|--offset=off     : Start offset in bytes\n"
          "                         (default 0)\n"
          "   -p|--pattern=pattern: Pattern of the file names to map\n"
          "                         (multiple pattern options are allowed)\n"
          "   -P|--pid=pid        : Pid of a process mapping the huge pages\n"
          "   -d|--debug=level    : Debug level\n"
          "   -F|--force          : Force same mapping (quite useless)\n"
          "   -h|--help           : This help\n"
          "\n"
          ,
          basename(pname)
         );

} // vmad_usage



// ----------------------------------------------------------------------------
// Name   : main
// Usage  : Program's entry point
// Return : None
// ----------------------------------------------------------------------------
int main(
         int   ac,
         char *av[]
        )
{
int           rc;
extern char  *optarg;
extern int    optind, opterr, optopt;
int           opt;
size_t        dump_offset = 0;
size_t        dump_size = 0;
const char  **pattern = NULL;
int           nb_patterns = 0;
int           nb_vma;
pid_t         tpid = -1;
vma_t        *vma;
size_t        dump_zone_size = 0;
int           i;
size_t        offset;
off_t         vma_off;
size_t        l;
link_t       *link;
int           force = 0;

  // Disable error messages from the library
  opterr = 0;

  while ((opt = getopt_long(ac, av, ":s:o:p:P:d:Fh", vmad_long_opts, NULL)) != EOF) {

    switch(opt) {

      case 's': {

        // Dump size
        dump_size = (size_t)strtoul(optarg, NULL, 10);

      }
      break;

      case 'o': {

        // Offset
        dump_offset = (size_t)strtoul(optarg, NULL, 10);

      }
      break;

      case 'p': {

        // Pattern
        nb_patterns ++;
        pattern = (const char **)realloc(pattern, sizeof(char *) * nb_patterns);
        if (!pattern) {

          ERR("realloc(): '%m' (%d)\n", errno);
          return 1;

        }

        pattern[nb_patterns - 1] = optarg;

      }
      break;

      case 'P': {

        // Pid of a process mapping the huge pages
        tpid = (pid_t)atoi(optarg);

      }
      break;

      case 'd': {

        // Debug level
        dbg_level = atoi(optarg);

      }
      break;

      case 'F': {

        // Force mapping to same virtual address of target pid
        force = 1;

      }
      break;

      case 'h': {

        // Help
        vmad_usage(av[0]);
        return 0;

      }
      break;

      case ':': {

        fprintf(stderr, "Missing value for option '%c'!\n", optopt);
        vmad_usage(av[0]);
        return 1;

      }
      break;

      case '?': {

        // If the option is known, there is an unexpected value for it
        // otherwise it is an unknown option
        switch(optopt) {

          case 'h': {

            fprintf(stderr, "Unexpected value for option '%c'!\n", optopt);

	  }
          break;

	  default: {

            if (optopt) {

              fprintf(stderr, "Unknown option '%c'!\n", optopt);

	    } else {

              fprintf(stderr, "Unknown option '%s'!\n", av[optind-1]);

	    }

	  }
          break;

	} // End switch

        vmad_usage(av[0]);
        return 1;

      }
      break;

      default: {
        vmad_usage(av[0]);
        return 1;
      }
      break;
    } // End switch

  } // End while

  if (tpid <= 0) {

    fprintf(stderr, "Missing target process identifier (-P option)!\n");
    vmad_usage(av[0]);
    return 1;

  }

  if (!pattern) {

    fprintf(stderr, "Missing file pattern (-p option)!\n");
    vmad_usage(av[0]);
    return 1;

  }

  // Initialize the list of VMAs
  vma_list.prev = vma_list.next = &vma_list;

  // Get the ordered list of VMA descriptors
  rc = vmad_map_pid(tpid, pattern, nb_patterns,  &nb_vma);
  if (0 != rc) {

    return 1;

  }

  // Map the files
  link = vma_list.next;
  while (link != &vma_list) {

    vma = LINK2VMA(link);

    // Open the file
    vma->fd = open(vma->fname, O_RDONLY);
    if (vma->fd < 0) {

      ERR("open(%s): '%m' (%d)\n", vma->fname, errno);
      return 1;

    }

    // Total dump zone size
    dump_zone_size += vma->len;

    // Map the file into memory
    vma->addr = mmap(vma->start_addr, vma->len, PROT_READ, MAP_SHARED, vma->fd, 0);
    if (MAP_FAILED == vma->addr) {

      ERR("mmap(%s, %zu): '%m' (%d)\n", vma->fname, vma->len, errno);
      return 1;

    }

    if (force && (vma->addr != vma->start_addr)) {

      fprintf(stderr, "Impossible to map '%s' at recommended address %p\n", vma->fname, vma->start_addr);
      return 1;

    }

    // Synchronize the data
    rc = msync(vma->addr, vma->len, MS_SYNC);
    if (rc < 0) {

      ERR("msync(%s@%p): '%m' (%d)\n", vma->fname, vma->addr, errno);
      return 1;

    }

    DBG(1, "Mmapped '%s' (fd=%d, size=%zu) @ %p (%s)\n",
        vma->fname,
        vma->fd,
        vma->len,
        vma->addr,
        (vma->addr == vma->start_addr ? "Same as source process" : "Different than source process"));

    link = link->next;

  } // End while

  // Dump size
  if (!dump_size) {

    dump_size = dump_zone_size - dump_offset;

  }

  if ((dump_zone_size - dump_offset) < dump_size) {

    fprintf(stderr, "Incoherent dump size (%zu) / offset (%zu) values\n", dump_size, dump_offset);
    return 1;

  }

  DBG(1, "Dumping %zu bytes from offset %zu\n", dump_size, dump_offset);

  // Dump
  offset = 0;
  link = vma_list.next;
  for (i = 0; (i < nb_vma) && dump_size; i ++) {

    vma = LINK2VMA(link);
    if ((offset + vma->len) > dump_offset) {

      l = dump_offset - offset;
      vma_off = (off_t)l;
      l = vma->len - l;
      if (l > dump_size) {

        l = dump_size;

      }

      DBG(2, "Dumping %zu bytes from offset %lu in '%s'\n",
          l, vma_off, vma->fname);

      rc = sendfile(1, vma->fd, &vma_off, l);
      if (rc != (int)l) {
        if (rc >= 0) {

          errno = EIO;

        }

        ERR("senfile(%lu, %zu): '%m' (%d)\n", vma_off, l, errno);
        return 1;

      }

      dump_offset += l;
      offset = dump_offset;
      dump_size -= l;

    } else {

      offset += vma->len;

    }

    link = link->next;

  } // End for

#if 0

  {
  int status;

    snprintf(fname, sizeof(fname), "cat /proc/%d/smaps", getpid());
    status = system(fname);

    printf("status = %d\n", status);
  }

  pause();

#endif // 0

  return 0;

} // main
